/*
	flounder - copyright 2002 Edward J. Vigmond

	This file is part of flounder.     
	
	flounder is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.     
	
	flounder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.     
	You should have received a copy of the GNU General Public License
    along with flounder; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "../config.h"
#include "flguppy.h"
#include <FL/math.h>
#include <cstring>

#ifdef HAVE_VALUES_H
#include <values.h>
#endif

#ifndef MAXDOUBLE
#define MAXDOUBLE 1.0e37
#endif
#ifndef MINDOUBLE
#define MINDOUBLE 1.0e-37
#endif

#include "contour.h"

const int Big_ENDIAN = 0;
const int Little_ENDIAN = 1;
#ifdef WORDS_BIGENDIAN
const int endian=Big_ENDIAN;
#else
const int endian=Little_ENDIAN;
#endif

const char *wintitle[6] = { "%s: XY(Z=%d,t=%d)", "%s: XZ(Y=%d,t=%d)",
                    "%s: YX(Z=%d,t=%d)", "%s: YZ(X=%d,t=%d)",
                    "%s: ZX(Y=%d,t=%d)", "%s: ZY(X=%d,t=%d)" };

const int     colourbarwidth=30;		// width of colour bar in pixels
const float   margin = 0.05;			// page margin for tableau in viewport units
const float   graphsep = 0.03;			// seperation for graphs in tableau
const GLfloat BGD_BLACK[3] = {0,0,1./255.};
const GLfloat BGD_WHITE[3] = {1,1,1.-1./255.};
const GLfloat mat_diffuse[] = { 0.7, 0.7, 0.7, 1.0 };
const GLfloat mat_diffuse_back[] = { 0.4, 0.4, 0.4, 1.0 };
const GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
const GLfloat mat_shininess[] = { 100.0 };
const GLfloat mat_shininess_back[] = { 80.0 };
const GLfloat difflight[] = { 1, 1, 1, 1 };
const float   light_dist=1;
const float   timeout=3.;				// close alert window in this time

void putText( char *, bool );
void windowhide( void *w );


// close an fl_alert window
void windowhide( void *w )
{
  if( ((Fl_Window *)w)->shown() ) Fl::handle(FL_CLOSE,(Fl_Window *)w);
}


// put tokens in the feedback buffer so an EPS file can be created which
// knows the actual characters printed.
// It puts the number of characters in the string before the gldraw() and
// then puts each character in the feedback buffer
void putText( char *s, bool feedback )
{
	if( feedback ) glPassThrough( float(strlen(s)) );
	gl_draw( s );
	if( feedback ) for( int i=0; i<strlen(s) ; i++ ) glPassThrough( float(s[i]) );
}


// return true if the pixel is defined to be transparent
bool Glarea::transparent( int x, int y, int z, int t )
{ 
	int a = header.x()*( y + header.y()*( z + header.z()*t ) ) + x;

	void* datp;
	float dat;
	switch( header.type() ){
		case IGB_BYTE:
			datp = ((unsigned char *)data)+a;
			dat = float(*((unsigned char *)datp));
			break;
		case IGB_CHAR:
			datp = ((signed char *)data)+a;
			dat = float(*((signed char *)datp));
			break;
		case IGB_SHORT:
			datp = ((short *)data)+a;
			dat = float(*((short *)datp));
			break;
		case IGB_LONG:
			datp = ((long *)data)+a;
			dat = float(*((long *)datp));
			break;
		case  IGB_FLOAT:
			datp = ((float *)data)+a;
			dat = float(*((float *)datp));
			break;
		case IGB_DOUBLE:
			datp = ((double *)data)+a;
			dat = float(*((double *)datp));
			break;
		case IGB_INT:
			datp = ((int *)data)+a;
			dat = float(*((int *)datp));
			break;
		case IGB_UINT:
			datp = ((unsigned int *)data)+a;
			dat = float(*((unsigned int *)datp));
			break;
		case IGB_RGBA:
			// check alpha channel
			return *(((char *)data)+4*a+3)==0;
		default:
			datp = NULL;
			dat  = 0;
	}
	if( header.transparent() != NULL &&
		    !memcmp( header.transparent(), datp, datumSize) )
			return 1;
	return (dat>=trans_min && dat<=trans_max) == !rngcomp; 
}

// the main heavy duty drawing routine
void Glarea::draw() 
{
	if (!valid()) {
		valid(1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPointSize( 10.0 );
		glPolygonMode( GL_FRONT, GL_FILL );
		glPolygonMode( GL_BACK, GL_LINE );
		glEnable( GL_CULL_FACE );
		glEnable( GL_LINE_SMOOTH );
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glShadeModel( GL_FLAT );
		glViewport(0,0,w(),h());
		glClearColor( bgdcolour[0], bgdcolour[1], bgdcolour[2], 0);
		if( numDim==2 )  {
			if( dispcolourbar )
				glOrtho( 0,  w()*xd/float(w()-colourbarwidth),  0, yd, -1,  1 );
			else 
				glOrtho( 0,  xd,  0, yd, -1,  1 );

		} else if( numDim==3 ) {
			glOrtho( -maxdim*0.8, maxdim*0.8, -maxdim*0.8, maxdim*0.8, 
					 -2.*maxdim, 2*maxdim );
			glRotatef( xrotate, 1, 0, 0 );
			glRotatef( yrotate, 0, 1, 0 );
			glRotatef( zrotate, 0, 0, 1 );

			GLfloat light_pos0[] = { header.x()/2, header.y()/2, header.z()*(1+zlightpos), 1.}; 
			glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
			GLfloat light_pos2[] = { (1+xlightpos)*header.x(), header.y()/2, header.z()/2, 1.}; 
			glLightfv( GL_LIGHT2, GL_POSITION, light_pos2 );
			GLfloat light_pos1[] = { header.x()/2, (1+ylightpos)*header.y(), header.z()/2, 0.}; 
			glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );
			GLfloat spotdir[] = { -1, 0, 0 };
			glLightfv( GL_LIGHT2, GL_SPOT_DIRECTION, spotdir );

			glLightfv( GL_LIGHT0, GL_DIFFUSE, difflight );
			//glLightfv( GL_LIGHT1, GL_DIFFUSE, difflight );
			glLightfv( GL_LIGHT2, GL_DIFFUSE, difflight );
			glLightfv( GL_LIGHT0, GL_SPECULAR, difflight );
			//glLightfv( GL_LIGHT1, GL_SPECULAR, difflight );
			glLightfv( GL_LIGHT2, GL_SPECULAR, difflight );

			//glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
			//glMaterialfv(GL_BACK, GL_DIFFUSE, mat_diffuse_back);
			glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
			glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
			//glMaterialfv(GL_BACK, GL_SHININESS, mat_shininess_back);

			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 0 );
			glColorMaterial( GL_FRONT_AND_BACK, GL_DIFFUSE );

			glEnable(GL_LIGHT0);
			//glEnable(GL_LIGHT1);
			glEnable(GL_LIGHT2);

			glTranslatef( -header.x()/2., -header.y()/2., -zscalefac*header.z()/2. );
			glScalef( 1, 1, zscalefac );
		}
	} 

	// determine colour mapping for current slice
	if( autocs == true ) calibratecolour();

	glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT );
	gl_font( fontface, fontsize );

	if( numDim == 2 ) {

		glDisable(GL_DEPTH_TEST);

		// draw the data
		if( showcontour==false || showdata == true ) {
		  if( !gouraudshade ){
			bool lasttrans=true;
			for( int i=0; i<xd; i++ ) {
			  glBegin( GL_QUAD_STRIP );
			  if( header.transparent() == NULL ||
					  memcmp(slice[i],header.transparent(),datumSize ) ){
				colourize(slice[i][0]);
				//else
				//glColor3fv( bgdcolour );
				glVertex2i( i+1, yd);
				glVertex2i( i, yd );
				lasttrans = false;
			  }
			  for( int j=0; j<yd; j++ ){
				if( header.transparent() == NULL ||
						memcmp(slice[i]+j,header.transparent(),datumSize ) ){
				  colourize(slice[i][j]);
				  //else
				  //glColor3fv( bgdcolour );
				  glVertex2i( i+1, yd-j-1 );
				  glVertex2i( i, yd-j-1 );
				  lasttrans = false;
				} else if( lasttrans == false  ){
				  glVertex2i( i+1, yd-j-1 );
				  glVertex2i( i, yd-j-1 );
				  glEnd();
				  lasttrans=true;
				  glBegin( GL_QUAD_STRIP );
				} 
			  }
			  glEnd();
			}
		  } else {
			glShadeModel( GL_SMOOTH );
			for( int i=0; i<xd; i++ ) {
			  glBegin( GL_QUAD_STRIP );
			  if( header.transparent() == NULL ||
					  memcmp(slice[i],header.transparent(),datumSize ) )
				colourize(slice[i][0]);
			  else
				glColor3fv( bgdcolour );
			  glVertex2i( i+1, yd );
			  if( header.transparent() == NULL ||
					  memcmp(slice[i],header.transparent(),datumSize ) )
				colourize(slice[i][0]);
			  else
				glColor3fv( bgdcolour );
			  glVertex2i( i, yd );
			  for( int j=0; j<yd; j++ ){
				if( header.transparent() == NULL ||
						memcmp(slice[i]+j,header.transparent(),datumSize ) )
				  colourize(slice[i][j]);
				else
				  glColor3fv( bgdcolour );
				glVertex2i( i+1, yd-j-1 );
				if( header.transparent() == NULL ||
						memcmp(slice[i]+j,header.transparent(),datumSize ) )
				  colourize(slice[i][j]);
				else
				  glColor3fv( bgdcolour );
				glVertex2i( i, yd-j-1 );
			  }
			  glEnd();
			}
			glShadeModel( GL_FLAT );
		  }
		}

		// draw contours
		if( showcontour==true ) {
			glDisable( GL_LINE_SMOOTH );
			glBegin(GL_LINES);
			for( int i=0; i<numcontour; i++ ) {
				Contour::generate_contour(slice,xd,yd,toraw(contourval+i*contourinc));

				if( showdata==true ) 
					glColor3fv( bgdcolour );
				else {
				  	float datval = toraw(contourval+i*contourinc);
					colourize(datval);
				}

				for( int j=0; j<Contour::num_vert; j++ )
					glVertex2f( Contour::vx[j]+0.5, yd-Contour::vy[j]-0.5 );
			}
			glEnd();
			glEnable( GL_LINE_SMOOTH );
		}

		/* draw colour scale bar */
		if( dispcolourbar ) {
			float xmax = w()*xd/float(w()-colourbarwidth);
			float xstart = xd+(xmax-xd)*0.25;
			glBegin( GL_QUADS );
			glColor3fv( bgdcolour );
			glVertex2f( xstart, 0 );
			glVertex2f( xstart, yd );
			glVertex2f( xd, yd );
			glVertex2f( xd, 0 );
			float cmap_size = float(cscale->size());
			for( int i=0; i<cmap_size; i++ ) {
				glColor3fv( cscale->entry(i) );
				glVertex2f( xmax,   float(yd)/cmap_size*i );
				glVertex2f( xmax,   float(yd)/cmap_size*(i+1) );
				glVertex2f( xstart, float(yd)/cmap_size*(i+1) );
				glVertex2f( xstart, float(yd)/cmap_size*i );
			}
			glEnd();
			float charwidth  = float(xd)/(w()-(dispcolourbar?colourbarwidth:0));
			float charheight = float(yd)/float(this->h());
			char text[256];
			glColor3fv( cursorColour );
			sprintf( text, "%.2g", minval() );
			glRasterPos2f( (w()-gl_width(text)-1)*charwidth, charheight );
			putText( text, feedingback );
			sprintf( text, "%.2g", maxval() );
			glRasterPos2f( (w()-gl_width(text)-1)*charwidth, 
								yd-(gl_height()-gl_descent())*charheight );
			putText( text, feedingback );
		}

	} else if( numDim ==3 ) { //3D

		glEnable(GL_DEPTH_TEST);

		// draw and label axes 
		if( disp3Daxes ) {
			glColor3fv( frameColour );
			glRasterPos3f( header.x()+maxdim*0.1, 0, 0 );
			putText( "X", feedingback );
			glRasterPos3f( 0, header.y()+maxdim*0.1, 0 );
			putText( "Y", feedingback );
			glRasterPos3f( 0, 0, header.z()+maxdim*0.1/zscalefac );
			putText( "Z", feedingback );
			glBegin( GL_LINES );
			glVertex3f( 0, 0, 0 );
			glVertex3f( header.x()+maxdim*0.1, 0, 0 );
			glVertex3f( 0, 0, 0 );
			glVertex3f( 0, header.y()+maxdim*0.1, 0 );
			glVertex3f( 0, 0, 0 );
			glVertex3f( 0, 0, header.z()+maxdim*0.1/zscalefac);
			glEnd();
		} 

		// turn on lights for 3D solids
		if( lightson == true ) glEnable(GL_LIGHTING);
		else glDisable(GL_LIGHTING);

		bool doiso = showiso[0];
		for( int i=1; i<numisosurf; i++ ) doiso = doiso || showiso[i];
		if( doiso ) {			// draw isosurfaces 
			Isosurf::DO_NORMALS = 1;
			//put time slice in proper form for isosurf routine
			Isosurf::alloc_3d_data_array(header.x(), header.y(), header.z());
			for( int i=0; i<header.z(); i++ )
				for( int j=0; j<header.y(); j++ )
					for( int k=0; k<header.x(); k++ )
						Isosurf::data[i][j][k] = getdatum(k,j,i,currTime);

			glDisable( GL_CULL_FACE );
			glEnable(GL_AUTO_NORMAL);
			glEnable(GL_NORMALIZE);
			if( isocol[0][3]!=1. || isocol[1][3]!=1. ) glEnable(GL_BLEND);
			glEnable(GL_COLOR_MATERIAL);
			glPolygonMode( GL_BACK, GL_FILL );

			for( int s=0; s<numisosurf; s++ ) {
				if( !showiso[s] ) continue;
				Isosurf::make_isosurface( header.x(), header.y() , header.z(), isovalue[s] );
				glColor4fv( isocol[s] );
				glBegin( GL_TRIANGLES );
				for( int i=0; i<Isosurf::num_vertices; i+=3 )  {
					float dx1=Isosurf::xvertices[i+1]-Isosurf::xvertices[i];
					float dy1=Isosurf::yvertices[i+1]-Isosurf::yvertices[i];
					float dz1=Isosurf::zvertices[i+1]-Isosurf::zvertices[i];
					float dx2=Isosurf::xvertices[i+2]-Isosurf::xvertices[i];
					float dy2=Isosurf::yvertices[i+2]-Isosurf::yvertices[i];
					float dz2=Isosurf::zvertices[i+2]-Isosurf::zvertices[i];
					glVertex3f( Isosurf::xvertices[i], Isosurf::yvertices[i], Isosurf::zvertices[i] );
					if( Isosurf::xnormals[i]*(dy1*dz2-dy2*dz1) +
						Isosurf::ynormals[i]*(dx2*dz1-dx1*dz2) +
						Isosurf::znormals[i]*(dx1*dy2-dx2*dy1)  > 0. ) {
						glVertex3f( Isosurf::xvertices[i+1], Isosurf::yvertices[i+1], Isosurf::zvertices[i+1] );
						glVertex3f( Isosurf::xvertices[i+2], Isosurf::yvertices[i+2], Isosurf::zvertices[i+2] );
					}else{
						glVertex3f( Isosurf::xvertices[i+2], Isosurf::yvertices[i+2], Isosurf::zvertices[i+2] );
						glVertex3f( Isosurf::xvertices[i+1], Isosurf::yvertices[i+1], Isosurf::zvertices[i+1] );
					}
				}
				glEnd();
			}
			Isosurf::free_3d_data_array(0);
					
			glDisable(GL_BLEND);
			glDisable(GL_NORMALIZE);
			glDisable(GL_AUTO_NORMAL);
			glDisable(GL_COLOR_MATERIAL);
			glEnable( GL_CULL_FACE );
			glPolygonMode( GL_BACK, GL_LINE );
			glDisable(GL_LIGHTING);
		} else if( show3Dsolid > 0 ) { 		// draw as a solid
			glEnable(GL_AUTO_NORMAL);
			glEnable(GL_NORMALIZE);
			glEnable(GL_COLOR_MATERIAL);

			// Louisiana method - simple to implement but bad results
			// quick to implement and dirty - just draw large points
			if( show3Dsolid == 1 ) {
				glBegin( GL_POINTS );
				for( int k=bbzm; k<bbzp; k++ ) 
					for( int j=bbym; j<bbyp; j++ ) 
						for( int i=bbxm; i<bbxp; i++ ){
							float datum = getdatum( i, j, k, currTime );
							if(!transparent(i, j, k, currTime ) ){
								colourize( datum );
								glVertex3f( i+0.5, j+0.5, k+0.5 );
							}
						}
				glEnd();
			} else if( show3Dsolid == 2 ) { //more pleasing 2 the eye
				// draw each pixel as a cube with 6 faces but
				// only draw a face if it can be seen
				glBegin( GL_QUADS );
				int* vlp=vis_vert_list;
				for( int i=0; i<num_visible_vert; i++, vlp+=15 ){
				  float datval;
				  getdatum(vlp[0], vlp[1], vlp[2], currTime, datval);
				  colourize(datval);
				  glVertex3iv( vlp+3 );
				  glVertex3iv( vlp+6 );
				  glVertex3iv( vlp+9 );
				  glVertex3iv( vlp+12 );
				}
				glEnd();
			}
			glDisable(GL_NORMALIZE);
			glDisable(GL_AUTO_NORMAL);
			glDisable(GL_COLOR_MATERIAL);
			glDisable(GL_LIGHTING);
							
		} else {
			// draw only bounding surfaces
			glDisable(GL_LIGHTING);
			DispDim olddispdim = dim_disp; //save current state
			dim_disp = DISP_XY;
			dataslice( currTime, bbzm );
			for( int i=bbxm; i<bbxp; i++ ){
			  glBegin( GL_QUAD_STRIP );
			  colourize(slice[i][bbyp-1]);
			  glVertex3i( i, bbyp, bbzm );
			  glVertex3i( i+1, bbyp, bbzm );
			  for( int j=bbyp-1; j>=bbym; j-- ){
				colourize(slice[i][j]);
				glVertex3i( i, j, bbzm );
				glVertex3i( i+1, j, bbzm );
			  }
			  glEnd();
			}
			dim_disp = DISP_XZ;
			dataslice( currTime, bbym );
			for( int i=bbxm; i<bbxp; i++ ){
			    glBegin( GL_QUAD_STRIP );
				colourize(slice[i][bbzp-1]);
				glVertex3i( i+1, bbym, bbzp );
				glVertex3i( i, bbym, bbzp );
				for( int j=bbzp-1; j>=bbzm; j-- ){
					colourize(slice[i][j]);
					glVertex3i( i+1, bbym, j );
					glVertex3i( i, bbym, j );
				}
			    glEnd();
			}
			dim_disp = DISP_YZ;
			dataslice( currTime, bbxm );
			for( int i=bbym; i<bbyp; i++ ){
			    glBegin( GL_QUAD_STRIP );
				colourize(slice[i][bbzp-1]);
				glVertex3i( bbxm, i, bbzp );
				glVertex3i( bbxm, i+1, bbzp );
				for( int j=bbzp-1; j>=bbzm; j-- ){
					colourize(slice[i][j]);
					glVertex3i( bbxm, i, j );
					glVertex3i( bbxm, i+1, j );
				}
			    glEnd();
			}
			dim_disp = DISP_XY;
			dataslice( currTime, bbzp-1 );
			for( int i=bbxm; i<bbxp; i++ ){
			    glBegin( GL_QUAD_STRIP );
				colourize(slice[i][bbyp-1]);
				glVertex3i( i+1, bbyp, bbzp );
				glVertex3i( i, bbyp, bbzp );
				for( int j=bbyp-1; j>=bbym; j-- ){
					colourize(slice[i][j]);
					glVertex3i( i+1, j, bbzp );
					glVertex3i( i, j, bbzp );
				}
			    glEnd();
			}
			dim_disp = DISP_XZ;
			dataslice( currTime, bbyp-1 );
			for( int i=bbxm; i<bbxp; i++ ){
			    glBegin( GL_QUAD_STRIP );
				colourize(slice[i][bbzp-1]);
				glVertex3i( i, bbyp, bbzp );
				glVertex3i( i+1, bbyp, bbzp );
				for( int j=bbzp-1; j>=bbzm; j-- ){
					colourize(slice[i][j]);
					glVertex3i( i, bbyp, j );
					glVertex3i( i+1, bbyp, j );
				}
				glEnd();
			}
			dim_disp = DISP_YZ;
			dataslice( currTime, bbxp-1 );
			for( int i=bbym; i<bbyp; i++ ){
			    glBegin( GL_QUAD_STRIP );
				colourize(slice[i][bbzp-1]);
				glVertex3i( bbxp, i+1, bbzp );
				glVertex3i( bbxp, i, bbzp );
				for( int j=bbzp-1; j>=bbzm; j-- ){
					colourize(slice[i][j]);
					glVertex3i( bbxp, i+1, j );
					glVertex3i( bbxp, i, j );
				}
				glEnd();
			}
			dim_disp = olddispdim;			// restore state
		}

		// now add code to draw the bounding box so we can see the edges 
		if( disp3Dframe ) {
			glColor3fv( frameColour );
			glDepthFunc(GL_LEQUAL);
			glDisable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT, GL_LINE);
			bool showisos = showiso[0] || showiso[1];

			glBegin(GL_QUADS);
			glVertex3i( 0, 0, 0);
			glVertex3i( 0, header.y(), 0);
			glVertex3i( header.x(), header.y(), 0);
			glVertex3i( header.x(), 0, 0);

			glVertex3i( 0, 0, header.z()-showisos );
			glVertex3i( header.x(), 0, header.z()-showisos );
			glVertex3i( header.x(), header.y(), header.z()-showisos );
			glVertex3i( 0, header.y(), header.z()-showisos );

			glVertex3i( 0, 0, 0);
			glVertex3i( header.x(), 0, 0);
			glVertex3i( header.x(), 0, header.z()-showisos);
			glVertex3i( 0, 0, header.z()-showisos);

			glVertex3i( 0, header.y(), 0);
			glVertex3i( 0, header.y(), header.z()-showisos);
			glVertex3i( header.x(), header.y(), header.z()-showisos);
			glVertex3i( header.x(), header.y(), 0);

			glVertex3i( 0, 0, 0);
			glVertex3i( 0, 0, header.z()-showisos);
			glVertex3i( 0, header.y(), header.z()-showisos);
			glVertex3i( 0, header.y(), 0);

			glVertex3i( header.x(), 0, 0);
			glVertex3i( header.x(), header.y(), 0);
			glVertex3i( header.x(), header.y(), header.z()-showisos);
			glVertex3i( header.x(), 0, header.z()-showisos);
			glEnd();

			glDepthFunc(GL_LESS);
			glPolygonMode( GL_FRONT, GL_FILL );
			glEnable(GL_CULL_FACE);
		}
	}

	// display value of time
	if( showtime==true ) {
		ostringstream tmstr;
		tmstr << "t = " << currTime*header.inc_t() + header.org_t() << '\0' ;
		glPushMatrix();
		glLoadIdentity();
		if( numDim == 2 )
			glColor3fv( cursorColour );
		else if( numDim == 3 )
			glColor3fv( frameColour );
		glRasterPos3f(  float(-0.95), float(-0.95) , 0 );
		char *c = strdup( tmstr.str().c_str() );
		putText( c, feedingback );
		free( c );
		glPopMatrix();
	}
}

// draw overlain items  N.B. they will not appear in any GIFs
void Glarea :: draw_overlay()
{
	if( numDim == 3 )  return;

	if( !valid() ) {
		valid(1);
		if( numDim==2 && dispcolourbar )
			glOrtho( 0,  w()*xd/float(w()-colourbarwidth),  0, yd, -1,  1 );
		else if( numDim==2 )
			glOrtho( 0,  xd,  0, yd, -1,  1 );
		else if( numDim==3 ) {
			glOrtho( -maxdim*0.8, maxdim*0.8, -maxdim*0.8, maxdim*0.8, 
					 -maxdim, maxdim );
			glRotatef( xrotate, 1, 0, 0 );
			glRotatef( yrotate, 0, 1, 0 );
			glRotatef( zrotate, 0, 0, 1 );
			glTranslatef( -header.x()/2., -header.y()/2., -zscalefac*header.z()/2. );
			glScalef( 1, 1, zscalefac );
		}
	}

	gl_font( fontface, fontsize );

	if( numDim == 2 ) {
		char text[256];
		// the following 2 variables convert pixels to ortho units 
		float charwidth  = float(xd)/( w()-(dispcolourbar?colourbarwidth:0) );
		float charheight = float(yd)/float(this->h());

		/* cursor display */
		if( dispcursor ) {
			float xcpos, ycpos;
			glColor3fv(cursorColour);

			// 0.5 added to go through centre of pixel
			if( displines ) {
				glBegin(GL_LINES);
				glVertex2f( xcursor+0.5, 0 ); 
				glVertex2f( xcursor+0.5, yd ); 
				glVertex2f( 0, 0.5+ycursor ); 
				glVertex2f( xd, 0.5+ycursor ); 
				glEnd();
			}

			if( dispvalue ) {
			    float datval;
				switch( dim_disp ) {
					case DISP_XY:
						getdatum( xcursor, yd-ycursor-1, currDepth, currTime, datval );
						break;
					case DISP_YZ:
						getdatum( currDepth, xcursor, yd-ycursor-1, currTime, datval );
						break;
					case DISP_ZX:
						getdatum( yd-ycursor-1, currDepth, xcursor, currTime, datval );
						break;
					case DISP_YX:
						getdatum( yd-ycursor-1, xcursor, currDepth, currTime, datval );
						break;
					case DISP_XZ:
						getdatum( xcursor, currDepth, yd-ycursor-1, currTime, datval );
						break;
					case DISP_ZY:
						getdatum( currDepth, yd-ycursor-1, xcursor, currTime, datval );
						break;
				}
				if( header.type() == IGB_RGBA ) {
				  int ip;
				  memcpy( &ip, &datval, sizeof(char)*4 );
				  uchar r=ip&0xff, g=(ip>>8)&0xff, b=(ip>>16)&0xff, a=(ip>>24)&0xff;
				  sprintf( text, "%.2x%.2x%.2x%.2x",  r, g, b, a );
				} else
					sprintf( text, "%.2g", toval( datval ) );
				if( xcursor > xd/2 ){
					xcpos = xcursor + 0.5 - (gl_width(text)+5)*charwidth;
				} else 
					xcpos = xcursor + 0.5 + 5*charwidth;
				if( ycursor > yd/2 ){
					ycpos = ycursor + 0.5 - (gl_height()-gl_descent())*charheight;
				} else 
					ycpos = ycursor + 0.5 + 5*charheight;
				gl_draw( text, xcpos, ycpos );
			}

			/* cursor x position */
			if( dispcoord ) {
				sprintf( text, "%d", xcursor );
				if( xcursor > xd/2 )
					xcpos = xcursor + 0.5 - (gl_width(text)+3)*charwidth;
				else
					xcpos = xcursor + 0.5 + 5*charwidth;
				if( ycursor > yd/2 )
					ycpos = gl_descent()*charheight;
				else 
					ycpos = yd - (gl_height()-gl_descent())*charheight;
				gl_draw( text, xcpos, ycpos );
				
				/* cursor y position */
				sprintf( text, "%d", ycursor );
				if( xcursor > xd/2 )
					xcpos = charwidth;
				else
					xcpos = xd - (5+gl_width(text))*charwidth;
				if( ycursor > yd/2 )
					ycpos = ycursor+0.5-(gl_height()-gl_descent())*charheight;
				else 
					ycpos = ycursor + 0.5 + 5*charheight;
				gl_draw( text, xcpos, ycpos );
			}
		}
		if( tableauoff == false ) {
			glPointSize(3.);
			glColor3fv( cursorColour );
			glBegin( GL_POINTS );
			for( float i=tabui->x0->value(); i<=tabui->xe->value(); i+=tabui->dx->value())
				for( float j=tabui->y0->value(); j<=tabui->ye->value(); j+=tabui->dy->value())
			  					
					glVertex2f( i+0.5, j+0.5 );
			glEnd();
			glPointSize(1.);
		}
	}
}


int  Glarea :: getdata( string fn )
{
	if( fn.size() == 0 ) {
		char *fnchar = fl_file_chooser( "Select an IGB file:", "*.igb*", "" );
		if( fnchar == NULL ) 
			return 1;
		else
			fn = fnchar;
	}

	gzFile in;
	fname = fn;
	if( (in = gzopen( fname.c_str(), "r" )) == NULL ) {
		/* try adding .igb to name */
		fname += ".igb";
		if( (in=gzopen( fname.c_str(), "r" )) == NULL ) {
			/* try adding a .gz */
			fname = fn + ".gz";
			if( (in=gzopen( fname.c_str(), "r" )) == NULL ) {
				/* try adding a .igb.gz */
				fname = fn + ".igb.gz";
				if( (in=gzopen( fname.c_str(), "r" )) == NULL ) 
					return 1;
			}
		}
	}

    /* read in header */
	header.fileptr( in );
    if( header.read() == 0 ) return 1 ;

	if( header.x()<1 || header.y()<1 || header.z()<1 || header.t()<1 ) return 1;

    /* determine byte order */
	int end;
    switch( header.systeme() ) {
        case DOS:
        case DOS3:
        case DOS4:
        case DOS5:
        case DOS6:
    	case OS2:
        case WINNT:
        case WINNTINTEL:
        case WINNTMIPS:
        case WINNTALPHA:
        case WINNTPOWER:
        case PC_LINUX:
		case IGB_LITTLE_ENDIAN:
            end = Little_ENDIAN;
            break;
		case IGB_BIG_ENDIAN:
        default:
            end = Big_ENDIAN;
    }
    
    switch( header.type() ) {
        case IGB_BYTE:
            datumSize = sizeof(unsigned char);
            break;
        case IGB_CHAR:
            datumSize = sizeof(signed char);
            break;
        case IGB_SHORT:
            datumSize = sizeof(short);
            break;
        case IGB_LONG:
            datumSize = sizeof(long);
            break;
        case IGB_FLOAT:
            datumSize = sizeof(float);
            break;
        case IGB_DOUBLE:
            datumSize = sizeof(double);
            break;
       	case IGB_INT:
            datumSize = sizeof(int);
            break;
        case IGB_UINT:
            datumSize = sizeof(unsigned int);
            break;
        case IGB_RGBA:
            datumSize = 4*sizeof(char);
            break;
        default:
            fprintf( stderr, "Type not recognized (yet)\n" );
            gzclose( in );
            return 1;
    }
	datalinks = new int;
	*datalinks = 1;
	unsigned int numdata = header.x()*header.y()*header.z()*header.t();
	data = malloc( numdata*datumSize );
	unsigned int bytes_read;
	if( (bytes_read=gzread(in, data, datumSize*numdata)) != numdata*datumSize ){
		ostringstream ost;
		ost << "Premature EOF reached\n" << fname << "\n" 
			<< bytes_read/datumSize/header.x()/header.y() << " slices of " <<
			header.z()*header.t()	<<  " read" << '\0' ;  
		Fl::add_timeout( timeout, windowhide, (void *)(fl_message_icon()->window()) );
		fl_alert( ost.str().c_str() );
		header.t(bytes_read/header.x()/header.y()/header.z()/datumSize);
		data = realloc( data, bytes_read*datumSize );
	} 
    gzclose( in );

    if( currTime>header.t() )
        currTime = header.t()-1;
    
    switch( dim_disp ) {
        case DISP_XY:
        case DISP_YX:
            if( currDepth>=header.z() )
                currDepth = header.z()-1;
            break;
        case DISP_XZ:
        case DISP_ZX:
            if( currDepth>=header.y() )
                currDepth = header.y()-1;
            break;
        case DISP_YZ:
        case DISP_ZY:
            if( currDepth>=header.x() ) currDepth = header.x()-1;
            break;
    }

    if( header.x()>header.y() && header.x()>header.z() )
        maxdim = header.x();
    else if( header.y()>header.z() )
        maxdim = header.y();
    else
        maxdim = header.z();

	// allocate 2D data matrix
    slice = (float **)malloc( maxdim*sizeof(float *) );
    for( int i=0; i<maxdim; i++ )
        slice[i] = (float *)malloc( maxdim*sizeof(float) );
    
    if( endian != end )
        swap_bytes();

	// determine short name to display in choice 
	menuname = fname;
	string::size_type i0 = menuname.rfind(".gz");
	if( i0 < string::npos) menuname.erase( i0, 3 );
	i0 = menuname.rfind(".igb");
	if( i0 < string::npos ) menuname.erase( i0, 4 );
	i0 = menuname.rfind("/");
	if( i0 != string::npos ) menuname = menuname.substr(i0+1,string::npos);

	bbxp = header.x();
	bbyp = header.y();
	bbzp = header.z();
	ycursor = header.y()-1;
	dataslice( 0, 0 );

	// determine data minimum and maximum
	float minimum = getdatum(0,0,0,0);
	float maximum = minimum;
	float dp;
	for( int i=0; i<header.x(); i++ )
		for( int j=0; j<header.y(); j++ )
			for( int k=0; k<header.z(); k++ )
				for( int t=0; t<header.t(); t++ ) {
					dp = getdatum( i, j, k, t );		
					if( dp < minimum ) minimum=dp;
					if( dp > maximum ) maximum=dp;
				}
	minval( toval( minimum ) );
	maxval( toval( maximum ) );

    return 0;
}


// try rereading file, assuming only the time has changed
int  Glarea :: reread()
{
	gzFile in;
	if( (in = gzopen( fname.c_str(), "rb" )) == NULL ) {
		string igbname = fname + ".igb";
		if( (in=gzopen( igbname.c_str(), "r" )) == NULL ) {
			igbname = fname + ".gz";
			if( (in=gzopen( igbname.c_str(), "r" )) == NULL ) {
				igbname = fname + ".igb.gz"; 
				if( (in=fopen( igbname.c_str(), "rb" )) == NULL ) 
					return 1;
			}
		}
	}

	// read in new header
	IGBheader* thead = new IGBheader(in);
    if( thead->read() == 0 ) {
        return 1 ;
    }
	if( header.x()<1 || header.y()<1 || header.z()<1 || header.t()<1 ) return 1;

	if( thead->x() != header.x() || thead->y() != header.y() || thead->z() != header.z() ||
		thead->type() != header.type()  								            ){
        fl_alert( "Reread ignored\npossible for changes in t only" );
		return 1;
	}
	header.t(thead->t());
    unsigned int numdata = header.x()*header.y()*header.z()*header.t();

	data = realloc( data, numdata*datumSize );
    unsigned int bytes_read;
    if( (bytes_read=gzread(in, data, datumSize*numdata)) != numdata*datumSize ){
		ostringstream ost;
		ost << "Premature EOF reached\n" << fname << "\n" 
				<< bytes_read/datumSize/header.x()/header.y() << " slices of " <<
			 header.z()*header.t()	<< " read" << '\0';  
		Fl::add_timeout( timeout, windowhide, (void *)(fl_message_icon()->window()) );
        fl_alert( ost.str().c_str() );
        header.t( bytes_read/header.x()/header.y()/header.z()/datumSize);
    }
    gzclose( in );
	delete thead;

    if( currTime>header.t() ) currTime = header.t()-1;
    
	int end;
    switch( header.systeme() ) {
        case DOS:
        case DOS3:
        case DOS4:
        case DOS5:
        case DOS6:
    	case OS2:
        case WINNT:
        case WINNTINTEL:
        case WINNTMIPS:
        case WINNTALPHA:
        case WINNTPOWER:
        case PC_LINUX:
            end = Little_ENDIAN;
            break;
        default:
            end = Big_ENDIAN;
    }
    if( endian != end )
        swap_bytes();

	dataslice( currTime, currDepth );
	redraw();
	update_time( currTime );
    return 0;
}


int Glarea::handle( int event ) 
{
	switch(event) {
		case FL_FOCUS:
			return 1;
		case FL_KEYBOARD:
			switch(Fl::event_key()) {
				case 'q': 
					if( Fl::event_state(FL_ALT) ) exit(0);
					return 1;
					break;
				case FL_Right:
					if( Fl::event_state(FL_SHIFT) )
						cursor(xcursor+1,ycursor);
					else
						pwin->gui->tmstep(1);
					return 1;
					break;
				case FL_Left:
					if( Fl::event_state(FL_SHIFT) )
						cursor(xcursor-1,ycursor);
					else
						pwin->gui->tmstep(-1);
					return 1;
					break;
				case FL_Up:
					if( Fl::event_state(FL_SHIFT) )
						cursor(xcursor,ycursor+1);
					else if( numDim == 2 ) {
						depth(currDepth+1);
						pwin->gui->depthslide->value(currDepth+1);
					}
					return 1;
					break;
				case FL_Down:
					if( Fl::event_state(FL_SHIFT) ) {
						cursor(xcursor,ycursor-1);
					    return 1;
					} else if( numDim == 2 ) {
						depth(currDepth-1);
						pwin->gui->depthslide->value(currDepth+1);
					    return 1;
					}
					break;
				case 'c':
					if( Fl::event_state(FL_SHIFT) )
					  pwin->gui->destroy_glwin(pwin);
					else
					  pwin->gui->window->show();
					return 1;
					break;
				case 'r':
					reread();
					pwin->gui->set_tm_max();
					return 1;
					break;
			}
			break;
		case FL_PUSH:
			if( Fl::event_state(FL_BUTTON1) ){
				pwin->gui->setcurrent(pwin);
				float xpos = xd*float(Fl::event_x())/
					float(w()-(dispcolourbar?colourbarwidth:0));
				cursor( int(xpos>=xd?xd-1:xpos), 
								int(float(h()-Fl::event_y())/float(h())*yd) );
				return 1;
			}
			break;
	}
	return 0;
}

// move the cursor and update auxialliary windows
void Glarea :: cursor( int nx, int ny )
{
	if( nx < xd && ny < yd && nx>=0 && ny>=0 ) {
		xcursor = nx;
		ycursor = ny;
		redraw_overlay();
		int n;
		double *t;
		double *d=timevec( n, t );
		pwin->timewin->set_data( n, d, currTime, t );
		d=horizontalvec( n );
		pwin->horizon->set_data( n, d );
		d=verticalvec( n );
		pwin->vertical->set_data( n, d );
	}
}


// Constructor for Glarea class
Glarea::Glarea (int X, int Y, int W, int H, const char *L):
 Fl_Gl_Window(X, Y, W, H, L)
{
	maxdim          = 0;
	slice           = NULL;
	autocs          = false;
	currDepth       = 0;
	currTime        = 0;
	disphoriz       = false;
	dispvert        = false;
	disptime        = false;
	dispcursor      = false;
	displines       = true;
	dispvalue       = true;
	dispcoord       = true;
	disptitle       = true;
	dispcolourbar   = false;
	xrotate         = -74;
	yrotate         = 0;
	zrotate         = -118;
	zscalefac       = 1;
	numDim          = 2;
	dim_disp        = DISP_XY;
	disp3Dframe     = true;
	disp3Daxes      = true;
	showtime        = false;
	bbxm = bbym = bbzm = 0;
	//mode(FL_ALPHA|FL_RGB|FL_DOUBLE|FL_DEPTH);
	mode(FL_RGB|FL_DOUBLE|FL_DEPTH);
	xcursor         = 0;
	memcpy( &bgdcolour, &BGD_BLACK, sizeof(bgdcolour) );
	frame_colour( FL_WHITE );
	cursor_colour( FL_WHITE );
	cscale	= new Colourscale;
	gouraudshade    = false;
	tableauoff      = true;
	fontface        = FL_HELVETICA_BOLD;
	fontsize        = 10;
	//isosurface stuff
	for( int s=0; s<numisosurf; s++ ){
		showiso[s]         = false;
		isovalue[s]        = 0.;
		isocol[s][3] 	   = 1.;
		isocol[s][1] 	   = 0.;
	}
	isocol[0][0] = 1.;
	isocol[1][2] = 1.;
	// contour stuff
	showcontour = showdata = false;
	numcontour = 1;
	contourval = contourinc = 0.;
	//3D solid modelling
	show3Dsolid = 0;
	trans_min = trans_max = 0.;
	// lights
	lightson = true;
	xlightpos = ylightpos = zlightpos = 2.;
	// feedback
	feedingback = false;
}


/* return a datum in float format */
float Glarea :: getdatum( int x, int y, int z, int t, float &datum )
{
    int a = header.x()*( y + header.y()*( z + header.z()*t ) ) + x;
    
    switch( header.type() ){
            case IGB_BYTE:
                datum = ((unsigned char *)data)[a];
                break;
            case IGB_CHAR:
                datum = ((signed char *)data)[a];
                break;
            case IGB_SHORT:
                datum = ((short *)data)[a];
                break;
            case IGB_LONG:
                datum = ((long *)data)[a];
                break;
            case  IGB_FLOAT:
                datum = ((float *)data)[a];
            	break;
            case IGB_DOUBLE:
                datum = ((double *)data)[a];
                break;
            case IGB_INT:
                datum = ((int *)data)[a];
                break;
            case IGB_UINT:
                datum = ((unsigned int *)data)[a];
                break;
            case IGB_RGBA:
                memcpy( &datum, ((unsigned char *)data)+4*a, 4*sizeof(char) );
                break;
            default:
                datum = 0;
        }
    return datum;
}


float Glarea :: getdatum( int x, int y, int z, int t )
{
  float datum;
  return getdatum( x, y, z, t, datum );
}
		  

// fill in an array with the current 2D slice raw data
void Glarea::dataslice( int time, int depth )
{
    switch( dim_disp ) {
        case DISP_XY:
            xd = header.x();
            yd = header.y();
            maxDepth = header.z();
            for( int i=0; i<xd; i++ )
                for( int j=0; j<yd; j++ )
                    getdatum( i, j, depth, time, slice[i][j] );
            break;
        case DISP_XZ:
            xd = header.x();
            yd = header.z();
     		maxDepth = header.y();
            for( int i=0; i<xd; i++ )
                for( int j=0; j<yd; j++ )
                    getdatum( i, depth, j, time, slice[i][j] );
            break;
        case DISP_YX:
            xd = header.y();
            yd = header.x();
            maxDepth = header.z();
            for( int i=0; i<xd; i++ )
                for( int j=0; j<yd; j++ )
                    getdatum( j, i, depth, time, slice[i][j] );
            break;
        case DISP_YZ:
            xd = header.y();
            yd = header.z();
            maxDepth = header.x();
            for( int i=0; i<xd; i++ )
                for( int j=0; j<yd; j++ )
                    getdatum( depth, i, j, time, slice[i][j] );
            break;
        case DISP_ZX:
            xd = header.z();
        	yd = header.x();
            maxDepth = header.y();
            for( int i=0; i<xd; i++ )
                for( int j=0; j<yd; j++ )
                    getdatum( j, depth, i, time, slice[i][j] );
            break;
        case DISP_ZY:
            xd = header.z();
            yd = header.y();
            maxDepth = header.x();
            for( int i=0; i<xd; i++ )
                for( int j=0; j<yd; j++ )
                    getdatum( depth, j, i, time, slice[i][j] );
            break;
    }
}


// select the colour scale
void Glarea :: colourscale( CScale_t cscalet )
{
	cscale->scale( cscalet ); redraw();
}

// set colour map to cover entire range of slice
void Glarea :: calibratecolour( void )
{
	double min, max;
	
	if( numDim == 2 ) {
		min = max = slice[0][0];
		for( int i=0; i<xd; i++ )
			for( int j=0; j<yd; j++ ){
				if( slice[i][j] > max ) max = slice[i][j];
				if( slice[i][j] < min ) min = slice[i][j];
			}
	} else {						// 3D
		min = max = getdatum( 0, 0, 0, currTime );
		for( int i=0; i<header.x(); i++ )
			for( int j=0; j<header.y(); j++ )
				for( int k=0; k<header.z(); k++ ) {
					float datum = getdatum( i, j, k, currTime );
					if( datum>max ) { max = datum; continue; }
					if( datum<min ) min = datum;
				}
	}
	cscale->calibrate( min, max );
}

void Glarea :: update_time( int t )
{
	if( t>=header.t() ) return;
	currTime = t;
	dataslice( currTime, currDepth );
	pwin->timewin->highlight( t );
	int n;
	double *d=horizontalvec( n );
	pwin->horizon->set_data( n, d );
	d=verticalvec( n );
	pwin->vertical->set_data( n, d );
	if( show3Dsolid && numDim==3 ) determine_visible_verts();
	redraw();
}

void Glarea :: numdim( const char *grouplab, const bool tableau_visible )
{
	if( !strcmp(grouplab," 2D ") || !strcmp( grouplab, "Contour")  ) {
		if( numDim != 2 ) {
			numDim = 2;
			/* make sure the window gets resized */
			DispDim cddim = dimdisp();
			if( cddim != DISP_XY )
				dim_disp = DISP_XY;
			else
				dim_disp = DISP_YZ;
			dimdisp( cddim );

			if( tableau_visible ) {
				tableauoff = false;
				redraw_overlay();
			}
			dataslice( currTime, currDepth );
		}
	} else if(!strcmp(grouplab," 3D ") || !strcmp(grouplab,"Isosurf") ||
								      	!strcmp(grouplab,"Solid 3D") 	) {
		if( numDim != 3 ) {
			numDim = 3;
			tableauoff = true;
			if( tableau_visible ) redraw_overlay();
			window()->size( 400, 400 );
			window()->size_range( 50, 50, 1000, 1000, 1, 1, 1);
		}
	} 
	valid(0);
	redraw();
}

void Glarea:: framedisp( int a ){ disp3Dframe = a; redraw(); }

// set colour of frame as specified by colour map index
void Glarea:: frame_colour( int colour )
{
	uchar r, g, b;
	Fl::get_color( Fl_Color(colour), r, g, b );
	flframecolour = colour;
	frameColour[0] = float(r)/255;
	frameColour[1] = float(g)/255;
	frameColour[2] = float(b)/255;
	redraw();
}
int Glarea:: frame_colour(){ return flframecolour; }

// set colour of cursor as specified by colour map index
void Glarea:: cursor_colour( int colour )
{
	uchar r, g, b;
	flcursorcolour = colour;
	Fl::get_color( Fl_Color(colour), r, g, b );
	cursorColour[0] = float(r)/255;
	cursorColour[1] = float(g)/255;
	cursorColour[2] = float(b)/255;
	redraw();
}
int Glarea:: cursor_colour(){ return flcursorcolour; }

void Glarea :: bgd( int w )
{
	if( w ){
		bgdcolour[0] = BGD_WHITE[0];
		bgdcolour[1] = BGD_WHITE[1];
		bgdcolour[2] = BGD_WHITE[2];
	}else {
		bgdcolour[0] = BGD_BLACK[0];
		bgdcolour[1] = BGD_BLACK[1];
		bgdcolour[2] = BGD_BLACK[2];
	}
	valid(0);
	redraw();
}
// return 1 if background is white
int Glarea :: bgd() { 
	return ( bgdcolour==BGD_WHITE )?1:0; 
}

void Glarea :: axes( int w )
{
	if( w )
		disp3Daxes = true;
	else
		disp3Daxes = false;
	valid(0);
	redraw();
}
int Glarea :: axes(){return disp3Daxes;}


void Glarea :: xrot( float r )
{
	xrotate = int(r);
	valid(0);
	redraw();
}


void Glarea :: yrot( float r )
{
	yrotate = int(r);
	valid(0);
	redraw();
}


void Glarea :: zrot( float r )
{
	zrotate = int(r);
	valid(0);
	redraw();
}

void Glarea :: zscale( float r )
{
	zscalefac = r;
	valid(0);
	redraw();
}


void Glarea :: update_bb( BBvalue dim, double lim )
{
	switch( dim ) {
		case BBXM: bbxm = int(lim)-1;
				break;
		case BBXP: bbxp = int(lim);
				break;
		case BBYM: bbym = int(lim)-1;
				break;
		case BBYP: bbyp = int(lim);
				break;
		case BBZM: bbzm = int(lim)-1;
				break;
		case BBZP: bbzp = int(lim);
				break;
	}
	if( show3Dsolid && numDim==3 ) determine_visible_verts();
	redraw();
}

int Glarea::get_bb( BBvalue dim )
{
	switch(dim) {
		case BBXM: return bbxm+1;
		case BBXP: return bbxp;
		case BBYM: return bbym+1;
		case BBYP: return bbyp;
		case BBZM: return bbzm+1;
		case BBZP: return bbzp;
	}
}


// change which 2 of 3 spatial dimensions are displayed
// if the specified dimension is the same as the current, do not resize
// the window
void Glarea :: dimdisp( DispDim d )
{
	currDepth = 0;
	bool resize = ( d != dim_disp );
	dim_disp = d;
	dataslice( currTime, currDepth );
	valid(0);
	//make sure to update the cursor and auxilliary windows
	if( xcursor>=xd ) xcursor=xd-1;
	if( ycursor>=yd ) ycursor=yd-1;
	cursor( xcursor, ycursor );

	if( resize == true ) { 
		double aspect = double(yd)/double(xd);
		double screenaspect = double(Fl::h())/double(Fl::w());
		if( aspect>=screenaspect )
			window()->size_range( 10, int(10*aspect), int(Fl::h()/aspect), 
																Fl::h(),1,1,1);
		else
			window()->size_range( int(10./aspect), 10, Fl::w(), 
											int(Fl::w()*aspect), 1,1,1);
		if( double(w())*aspect<=Fl::w() )
			window()->size( w(), int(w()*aspect) );
		else
			window()->size(int(double(h())/aspect),h());
	}
	redraw();
}
DispDim Glarea :: dimdisp(){ return dim_disp; }

void Glarea :: depth( int d )
{
	if( d>=0 && d<maxDepth ) {
		currDepth = d;
		dataslice( currTime, currDepth );
		redraw();
	}
}
int Glarea :: depth(){ return currDepth; }

// set minimum data value
void Glarea :: minval( double min )
{
	cscale->calibrate( toraw(min), cscale->max() );
	redraw();
}

//return minimum data value displayed
double Glarea :: minval( void )
{
	return toval(cscale->min());
}

// set maximum value for colour scale given maximum data value
void Glarea :: maxval( double max )
{
	cscale->calibrate( cscale->min(), toraw(max) );
	redraw();
}

// return maximum data value displayed in data units
double Glarea :: maxval( void )
{
	return toval(cscale->max());
}

void Glarea :: autocol( bool a )
{
	autocs = a;
	redraw();
}
bool Glarea :: autocol(){ return autocs; }

void Glarea :: cursor( bool a )
{
	dispcursor = a;
	redraw_overlay();
}
bool Glarea :: cursor(){ return dispcursor; }

void Glarea:: curlines( bool a )
{
	displines = a;
	redraw_overlay();
}
bool Glarea:: curlines(){ return displines; }

// control display of cursor value
void Glarea :: curval( bool a )
{
	dispvalue = a;
	redraw_overlay();
}
bool Glarea :: curval(){ return dispvalue; }

// display cursor coordinates
void Glarea :: curcoord( bool a )
{
	dispcoord = a;
	redraw_overlay();
}
bool Glarea :: curcoord(){ return dispcoord; }

// control display of colour bar
void Glarea :: 	colourbar( bool a )
{
	dispcolourbar = a;
	valid(0);
	redraw();
}
bool Glarea :: 	colourbar(){ return dispcolourbar; }


double *Glarea :: horizontalvec( int &n )
{
    static double *vec=NULL;
	n = xd;
	vec = (double *)realloc((void *)vec, xd*sizeof(double) );
	for( int i=0;i<xd; i++ )
		vec[i] = toval(slice[i][yd-1-ycursor]);
	return vec;
}

	
double *Glarea :: verticalvec( int &n )
{
    static double *vec=NULL;
	n = yd;	
	vec = (double *)realloc((void *)vec, yd*sizeof(double) );
	for( int i=0;i<yd; i++ )
		vec[i] = toval(slice[xcursor][yd-1-i]);
	return vec;
}


double *Glarea :: timevec( int &n, double* &t )
{
    static double *vec=NULL, *tm=NULL;
	
	vec = (double *)realloc((void *)vec, header.t()*sizeof(double) );
	tm  = (double *)realloc((void *)tm, header.t()*sizeof(double) );

	n = header.t();
    for( int i=0; i<header.t(); i++ ){
		tm[i] = header.org_t() + i*header.inc_t();
        switch( dim_disp ) {
            case DISP_XY:
                vec[i] = getdatum( xcursor, yd-ycursor-1, currDepth, i );
                break;
            case DISP_XZ:
                vec[i] = getdatum( xcursor, currDepth, yd-ycursor-1, i );
                break;
            case DISP_YX:
                vec[i] = getdatum( yd-ycursor-1, xcursor, currDepth, i );
                break;
            case DISP_YZ:
                vec[i] = getdatum( currDepth, xcursor, yd-ycursor-1, i );
                break;
            case DISP_ZX:
                vec[i] = getdatum( yd-ycursor-1, currDepth, xcursor, i );
   				 break;
            case DISP_ZY:
                vec[i] = getdatum( currDepth, yd-ycursor-1, xcursor, i );
                break;
        }
		vec[i] = toval(vec[i]);
	}
	t = tm;
    return vec;
}


Glarea::~Glarea()
{
	if( !(*datalinks)-- ) free( data );
	for( int i=0; i<maxdim; i++ )
		free( slice[i] );
	free( slice );
}

void Glarea::shading( bool a )
{
	gouraudshade = a;
	redraw();
}

bool Glarea::shading(){ return gouraudshade; }

// share data from another view
void Glarea :: copy( Glarea *b )
{
	header = b->header;
	data   = b->data;
	(*(datalinks = b->datalinks))++;
	cscale->calibrate( toraw(b->minval()), toraw(b->maxval()) );
	fname = b->fname;
	currTime = b->currTime;
	maxdim = b->maxdim;
	menuname = b->menuname;
	bbxp = header.x();
	bbyp = header.y();
	bbzp = header.z();
    slice = (float **)malloc( maxdim*sizeof(float *) );
    for( int i=0; i<maxdim; i++ )
        slice[i] = (float *)malloc( maxdim*sizeof(float) );
	ycursor = header.y()-1;
	dataslice( currTime, 0 );
}
    
	
void Glarea :: swap_bytes()
{
    unsigned char tmpb;
    
    if( datumSize==1 || header.type()==IGB_RGBA )
        return;

    unsigned char *bp = (unsigned char *)data;
    int nd = header.x()*header.y()*header.z()*header.t();
    switch( datumSize ) {
        case 2:
            for( int i=0; i<nd; i++ ) {
                tmpb = bp[0];
                bp[0] = bp[1];
                bp[1] = tmpb;
                bp += datumSize;
            }
            break;
        case 4:
            for( int i=0; i<nd; i++ ) {
                tmpb = bp[0];
                bp[0] = bp[3];
        		bp[3] = tmpb;
                tmpb = bp[1];
                bp[1] = bp[2];
                bp[2] = tmpb;
                bp += datumSize;
            }
            break;
        case 8:
            for( int i=0; i<nd; i++ ) {
                tmpb = bp[0];
                bp[0] = bp[7];
                bp[7] = tmpb;
                tmpb = bp[1];
                bp[1] = bp[6];
                bp[6] = tmpb;
                tmpb = bp[2];
                bp[2] = bp[5];
                bp[5] = tmpb;           
                tmpb = bp[3];
                bp[3] = bp[4];
                bp[4] = tmpb;           
                bp += datumSize;
            }
   			break;  
    }
	dataslice( currTime, currDepth );
}

// update the tableau user interface widget
void Glarea :: tableau( TableauUI *tb, bool update ) 
{
	if( tb->window->visible() == false ) return;
	
	tableauoff = false;
	tb->x0->minimum(0);
	tb->x0->maximum(xd-1);
	tb->x0->value(0);
	tb->xe->minimum(0);
	tb->xe->maximum(xd-1);
	tb->xe->value(xd-1);
	tb->dx->value(1);
	tb->y0->minimum(0);
	tb->y0->maximum(yd-1);
	tb->y0->value(0);
	tb->ye->minimum(0);
	tb->ye->maximum(yd-1);
	tb->ye->value(yd-1);
	tb->dy->value(1);
	tb->t0->minimum(0);
	tb->t0->maximum(header.t()-1);
	tb->t0->value(0);
	tb->te->minimum(0);
	tb->te->maximum(header.t()-1);
	tb->te->value(header.t()-1);
	tb->dt->value(1);
	tb->window->show();
	tb->widg = (void *)this;
	tb->update();
}

/* 
 * read the tableau user interface widget and open a Grace
 * process to graph the time series of the specified pixels
 */
void Glarea :: plot_tableau( TableauUI *u )
{
#ifndef HAVE_LIBGRACE_NP 
	fprintf( stderr, "**** Not implemented since grace is missing ****\n" );
	return;
#else
	int xcurs = xcursor;			// store current cursor position
	int ycurs = ycursor;
	
	int numx = int((u->xe->value()-u->x0->value())/u->dx->value()) + 1;
	int numy = int((u->ye->value()-u->y0->value())/u->dy->value()) + 1;
	int numt = int((u->te->value()-u->t0->value())/u->dt->value()) + 1;

	if( numx <= 0 || numy <= 0 ||  numt<=1 ){
		if( numx <= 0 ) fl_alert( "Illegal x dimension ( <1 )" );
		if( numy <= 0 ) fl_alert( "Illegal y dimension ( <1 )" );
		if( numt <= 1 ) fl_alert( "Illegal t dimension ( <=1 )" );
		return;
	}

	double ymin = MAXDOUBLE;
	double ymax = MINDOUBLE;
	int    gnum = -1;
	int    column=0;
	double xsep = graphsep/sqrt(double(numx));
	double ysep = graphsep/sqrt(double(numy));
	float  graphwidth  = (1.-2*margin-(numx-1)*xsep)/float(numx);
	float  graphheight = (1.-2*margin-(numy-1)*ysep)/float(numy);
	GraceOpenVA( "xmgrace", 2048, "-barebones", 
			"-geometry", "606X637+0+0", "-fixed", "600", "600", NULL );
	for( double i=u->x0->value(); i<=u->xe->value(); i+= u->dx->value() ) {
		int row = 0;
		for( double j=u->y0->value(); j<=u->ye->value(); j+= u->dy->value() ) {
			gnum++;
			int n;
			xcursor = int(i);
			ycursor = int(j);
			double *t;
			double *d = timevec( n, t );
			GracePrintf("with g%d", gnum );
			GracePrintf("xaxis off");
			GracePrintf("yaxis off");
			GracePrintf("world xmin %f", 
						u->t0->value()*header.inc_t()+header.org_t() );
			GracePrintf("world xmax %f", 
						u->te->value()*header.inc_t()+header.org_t() );
			GracePrintf("view xmin %f", margin+column*(xsep+graphwidth) );
			GracePrintf("view xmax %f", margin+column*xsep+(column+1)*graphwidth);
			GracePrintf("view ymin %f", margin+row*(ysep+graphheight) );
			GracePrintf("view ymax %f", margin+row*ysep+(row+1)*graphheight);
			for( int k=int(u->t0->value()); k<=u->te->value(); k+=int(u->dt->value()) ){
				if( d[k] > ymax ) ymax = d[k];
				if( d[k] < ymin ) ymin = d[k];
				GracePrintf( "s0 point %g, %g", k*header.inc_t()+header.org_t(), d[k] );
			}
			row++;
		}
		column++;
	}
	for( ; gnum >= 0; gnum-- ) {
		GracePrintf("with g%d", gnum );
		GracePrintf("world ymin %f", ymin );
		GracePrintf("world ymax %f", ymax );
		GracePrintf("frame off" );
	}
	GracePrintf( "redraw" );
	GraceClosePipe();
	xcursor = xcurs;
	ycursor = ycurs;
#endif
}

void Glarea :: tableau_area( TableauUI *tb )
{
  tabui = tb;
  redraw_overlay();
}

void Glarea :: levels( int numlev )
{
	if( numlev<2 || numlev>252 ) return;
	cscale->size( numlev );
	redraw();
}


// copy the current data colour map into the GIF colormapobject
// return the size of the colour map
//
int Glarea :: gif_cmap( ColorMapObject *gifcmap )
{
	// frame, axes, text colours
	int coff=0;
	gifcmap->Colors[coff].Red     = int(rint(bgdcolour[0]*255));
	gifcmap->Colors[coff].Green   = int(rint(bgdcolour[1]*255));
	gifcmap->Colors[coff++].Blue  = int(rint(bgdcolour[2]*255));
	gifcmap->Colors[coff].Red     = int(rint(cursorColour[0]*255));
	gifcmap->Colors[coff].Green   = int(rint(cursorColour[1]*255));
	gifcmap->Colors[coff++].Blue  = int(rint(cursorColour[2]*255));
	gifcmap->Colors[coff].Red     = int(rint(frameColour[0]*255));
	gifcmap->Colors[coff].Green   = int(rint(frameColour[1]*255));
	gifcmap->Colors[coff++].Blue  = int(rint(frameColour[2]*255));

	if( showiso[0] || showiso[1] ) { 	 // show isosurfaces
		int i=coff;
		const float intvl = 59; 	     // (#colours-1) in ramp
		const float frameintvl = 15; 	 // (#colours-1) in frame-iso ramp
		
		//set up colour map for alpha blending

		// merge isosurface colours with background
		for( int k=0; k<numisosurf; k++ )
			for( float j=0; j<=intvl; j++, i++ ) {
				gifcmap->Colors[i].Red   = int(rint( 255*(
					j/intvl*(isocol[k][0]-bgdcolour[0])+bgdcolour[0])));
				gifcmap->Colors[i].Green = int(rint( 255*(
					j/intvl*(isocol[k][1]-bgdcolour[1])+bgdcolour[1])));
				gifcmap->Colors[i].Blue  = int(rint( 255*(
					j/intvl*(isocol[k][2]-bgdcolour[2])+bgdcolour[2])));
			}
		// merge iso0 with iso1
		for( float j=0; j<=intvl; j++, i++ ) {
			gifcmap->Colors[i].Red   = int(rint( 255*(
				j/intvl*(isocol[1][0]-isocol[0][0])+isocol[0][0])));
			gifcmap->Colors[i].Green = int(rint( 255*(
				j/intvl*(isocol[1][1]-isocol[0][1])+isocol[0][1])));
			gifcmap->Colors[i].Blue  = int(rint( 255*(
				j/intvl*(isocol[1][2]-isocol[0][2])+isocol[0][2])));
		}
		// merge isosurface colours with frame 
		for( int k=0; k<numisosurf; k++ )
			for( float j=0; j<=frameintvl; j++, i++ ) {
				gifcmap->Colors[i].Red   = int(rint( 255*(
					j/intvl*(isocol[k][0]-frameColour[0])+frameColour[0])));
				gifcmap->Colors[i].Green = int(rint( 255*(
					j/intvl*(isocol[k][1]-frameColour[1])+frameColour[1])));
				gifcmap->Colors[i].Blue  = int(rint( 255*(
					j/intvl*(isocol[k][2]-frameColour[2])+frameColour[2])));
			}
		return i;
	} else {
		for( int i=0; i<cscale->size(); i++ ){
			GLfloat* cvec=cscale->entry(i);
			gifcmap->Colors[i+coff].Red   = int(rint(cvec[0]*255));
			gifcmap->Colors[i+coff].Green = int(rint(cvec[1]*255));
			gifcmap->Colors[i+coff].Blue  = int(rint(cvec[2]*255));
		}
		return cscale->size()+coff;
	}
}

// display the current time
void Glarea :: showtm( bool a )
{
	showtime = a;
	redraw();
}

void Glarea :: change_font( Fl_Font fontid, int size )
{
	if( size<4 || size>100 ) return;
	fontface = fontid;
	fontsize = size;
	redraw();
}

// set isosurface colour
void Glarea :: isosurf_colour( int s, double r, double g, double b, double a )
{
	isocol[s][0] = GLfloat(r);
	isocol[s][1] = GLfloat(g);
	isocol[s][2] = GLfloat(b);
	isocol[s][3] = GLfloat(a);
	redraw();
}
GLfloat *Glarea :: isosurf_colour( int s ){ return isocol[s]; }

// get/set contour values
// set update to return the values into the vars, otherwise
// set the variables from the supplied parameters
void Glarea :: contours( bool& con, bool& dat, int& nc, double& st, double& inc, 
									bool update )
{
	if( update == false ) {
		showcontour = con;
		showdata    = dat;
		numcontour  = nc;
		contourval  = st;
		contourinc  = inc;
		redraw();
	} else {
		con = showcontour;
		dat = showdata;
		nc  = numcontour;
		st  = contourval;
		inc = contourinc;
	}
}


//call GUI to down sample the data
void Glarea:: subsample()
{
	SubsampleUI *ssui = new SubsampleUI;
	ssui->xstrslide->setnum(header.x());
	ssui->ystrslide->setnum(header.y());
	ssui->zstrslide->setnum(header.z());
	ssui->get_strides();
	int xstr = ssui->xstrslide->value();
	int ystr = ssui->ystrslide->value();
	int zstr = ssui->zstrslide->value();
	int nx   = ( header.x() + xstr -1 )/xstr;
	int ny   = ( header.y() + ystr -1 )/ystr;
	int nz   = ( header.z() + zstr -1 )/zstr;

	void* data1 = malloc(nx*ny*nz*header.t()*datumSize);
	for( int t=0; t<header.t(); t++ )
		for( int z=0; z<nz; z++ )
			for( int y=0; y<ny; y++ )
				for( int x=0; x<nx; x++ )
					memcpy( 
						(char *)data1+(((t*nz + z)*ny + y)*nx + x)*datumSize,
						(char *)data+(((t*header.z() + z*zstr)*header.y() + 
									     y*ystr)*header.x()	+ x*xstr)*datumSize,
																datumSize );
	free( data );
	data = data1;

	// now adjust things that have changed 
	header.x(nx);
	header.y(ny);
	header.z(nz); 
	header.inc_x(header.inc_x() * xstr);
	header.inc_y(header.inc_y() * ystr);
	header.inc_z(header.inc_z() * zstr);
	header.fac_x(header.fac_x() * xstr);
	header.fac_y(header.fac_y() * ystr);
	header.fac_z(header.fac_z() * zstr);

    if( header.x()>header.y() && header.x()>header.z() )
        maxdim = header.x();
    else if( header.y()>header.z() )
        maxdim = header.y();
    else
        maxdim = header.z();

	bbxp = header.x();
	bbyp = header.y();
	bbzp = header.z();

	dimdisp( dim_disp );  //this will do a redraw

	if( xcursor>=xd ) xcursor = xd/2;
	if( ycursor>=yd ) ycursor = yd/2;
	redraw_overlay();
}

// set the transparency range for 3D solid modelling
void Glarea:: transrange( double min, double max, bool range )
{
	trans_min = toraw(min);
	if( range == true ) trans_max = toraw( max );
	else trans_max = trans_min;
	if( show3Dsolid ) {
	  	determine_visible_verts();
		redraw();
	}
}

// get the transparency range
bool Glarea::transrange( float& min, float &max )
{
	min=toval(trans_min);
	max=toval(trans_max); 
	return min != max;
}


// set light position
void Glarea:: lightpos( char light, float pos )
{
	switch( light ) {
		case 'X':
			xlightpos = pos;
			break;
		case 'Y':
			ylightpos = pos;
			break;
		case 'Z':
			zlightpos = pos;
			break;
	}
	valid(0);
	redraw();
}

// get light postion
float Glarea:: lightpos( char light )
{
	switch( light ) {
		case 'X': return xlightpos;
		case 'Y': return ylightpos;
		case 'Z': return zlightpos;
	}
}

// render as an EPS object
void Glarea :: renderAsEPS( RenderEPS *r )
{
  int returned;
  r->buffersize( xd*yd*50 );
  do {
	glFeedbackBuffer( r->buffersize(), GL_3D_COLOR, r->feedbackbuf() );
	glRenderMode(GL_FEEDBACK);
	feedingback = true;
	draw();
	feedingback = false;
	returned = glRenderMode(GL_RENDER);
	if( returned<0 )
	  r->buffersize( 10*r->buffersize() );
  }while( returned<0 );
  GLfloat vec[4];
  glGetFloatv(GL_VIEWPORT, vec);
  r->viewprt( vec );
  glGetFloatv(GL_COLOR_CLEAR_VALUE, vec);
  r->clearcolor( vec );
  r->linewidth( 2 );
  r->pointsize( 3 );
  r->fontsize( float(fontsize) );
  switch( fontface ){
	  case FL_HELVETICA:
		  r->font( "Helvetica" );
		  break;
	  case FL_HELVETICA_BOLD:
		  r->font( "Helvetica-Bold" );
		  break;
	  case FL_HELVETICA_ITALIC:
		  r->font( "Helvetica-Italic" );
		  break;
	  case FL_HELVETICA_BOLD_ITALIC:
		  r->font( "Helvetica-Bold-Italic" );
		  break;
	  case FL_COURIER:
		  r->font( "Courier" );
		  break;
	  case FL_COURIER_BOLD:
		  r->font( "Courier-Bold" );
		  break;
	  case FL_COURIER_ITALIC:
		  r->font( "Courier-Italic" );
		  break;
	  case FL_COURIER_BOLD_ITALIC:
		  r->font( "Courier-Bold-Italic" );
		  break;
	  case FL_TIMES:
		  r->font( "Times-Roman" );
		  break;
	  case FL_TIMES_BOLD:
		  r->font( "Times-Roman-Bold" );
		  break;
	  case FL_TIMES_ITALIC:
		  r->font( "Times-Roman-Italic" );
		  break;
	  case FL_TIMES_BOLD_ITALIC:
		  r->font( "Times-Roman-Bold-Italic" );
		  break;
	  case FL_SYMBOL:
		  r->font( "Symbol" );
		  break;
	  case FL_SCREEN:
		  r->font( "AvanteGarde-Book" );
		  break;
	  case FL_SCREEN_BOLD:
		  r->font( "AvanteGarde-Demi" );
		  break;
  }
  if( numDim==3 || dispcolourbar  )
	r->fillBackground( fl_ask("Do you want the background filled?") );
  r->render( returned );
}

void Glarea :: solid3D( int a )
{
  show3Dsolid=a;
  if( a ) determine_visible_verts();
  redraw();
}

//determine the visible vertices for solid 3D rendering
void Glarea:: determine_visible_verts(void)
{
  num_visible_vert = 0;
  free( vis_vert_list );
  vis_vert_list = NULL;
	  
  char*** vl = new char**[bbxp-bbxm];
  vl -= bbxm;
  for( int i=bbxm; i<bbxp; i++ ){
	  vl[i] = new char*[bbyp-bbym];
	  vl[i] -= bbym;
	  for( int j=bbym; j<bbyp; j++ ){
		  vl[i][j] = new char [bbzp-bbzm];
		  vl[i][j] -= bbzm;
		  for( int k=bbzm; k<bbzp; k++ )
			  vl[i][j][k] = transparent(i, j, k, currTime);
	  }
  }
  for( int k=bbzm; k<bbzp; k++ ) 
	for( int j=bbym; j<bbyp; j++ ) 
	  for( int i=bbxm; i<bbxp; i++ ){
		if( vl[i][j][k] /*transparent( i, j, k, currTime)*/ ) continue;
		float datval;
		getdatum( i, j, k, currTime, datval );
		colourize( datval );
		// look at previous cube in x direction
		if( i==bbxm || vl[i-1][j][k] ) {
				//transparent(i-1, j, k, currTime) ) {
		  vis_vert_list = (int *)realloc( vis_vert_list, 
				  						15*++num_visible_vert*sizeof(int) );
		  int* p = vis_vert_list + (num_visible_vert-1)*15;
		  *(p++) = i;				// yz
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;
		  *(p++) = j;
		  *(p++) = k+1;
		  *(p++) = i;
		  *(p++) = j+1;
		  *(p++) = k+1;
		  *(p++) = i;
		  *(p++) = j+1;
		  *(p++) = k;
		}
		// look at previous cube in y direction
		if( j==bbym || vl[i][j-1][k] ) {
				//transparent(i, j-1, k, currTime) ) {
		  vis_vert_list = (int *)realloc( vis_vert_list, 
				  						15*++num_visible_vert*sizeof(int) );
		  int* p = vis_vert_list + (num_visible_vert-1)*15;
		  *(p++) = i;				// zx
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j;
		  *(p++) = k+1;
		  *(p++) = i;
		  *(p++) = j;
		  *(p++) = k+1;
		}
		// look at previous cube in z direction
		if( k==bbzm || vl[i][j][k-1] ) {
				//transparent(i, j, k-1, currTime) ) {
		  vis_vert_list = (int *)realloc( vis_vert_list, 
				  						15*++num_visible_vert*sizeof(int) );
		  int* p = vis_vert_list + (num_visible_vert-1)*15;
		  *(p++) = i;				// xy
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;		
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;
		  *(p++) = j+1;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j+1;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j;
		  *(p++) = k;
		}
		// try next cube in x direction
		if( i==bbxp-1 || vl[i+1][j][k] ) {
		//		transparent(i+1, j, k, currTime) ) {
		  vis_vert_list = (int *)realloc( vis_vert_list, 
				  						15*++num_visible_vert*sizeof(int) );
		  int* p = vis_vert_list + (num_visible_vert-1)*15;
		  *(p++) = i;				// zy
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j+1;
		  *(p++) = k;
		  *(p++) = i+1;
		  *(p++) = j+1;
		  *(p++) = k+1;
		  *(p++) = i+1;
		  *(p++) = j;
		  *(p++) = k+1;
		}
		// try next cube in y direction
		if( j==bbyp-1 || vl[i][j+1][k] ) {
				//transparent( i, j+1, k, currTime ) ) {
		  vis_vert_list = (int *)realloc( vis_vert_list, 
				  						15*++num_visible_vert*sizeof(int) );
		  int* p = vis_vert_list + (num_visible_vert-1)*15;
		  *(p++) = i;				// xz
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;	
		  *(p++) = j+1;
		  *(p++) = k;
		  *(p++) = i;
		  *(p++) = j+1;
		  *(p++) = k+1;
		  *(p++) = i+1;
		  *(p++) = j+1;
		  *(p++) = k+1;
		  *(p++) = i+1;
		  *(p++) = j+1;
		  *(p++) = k;
		}
		// try next cube in z direction
		if( k==bbzp-1 || vl[i][j][k+1] ) {
		//		transparent( i, j, k+1, currTime) ) {
		  vis_vert_list = (int *)realloc( vis_vert_list, 
				  						15*++num_visible_vert*sizeof(int) );
		  int* p = vis_vert_list + (num_visible_vert-1)*15;
		  *(p++) = i;				// yx
		  *(p++) = j;
		  *(p++) = k;
		  *(p++) = i;		
		  *(p++) = j;
		  *(p++) = k+1;
		  *(p++) = i+1;
		  *(p++) = j;
		  *(p++) = k+1;
		  *(p++) = i+1;
		  *(p++) = j+1;
		  *(p++) = k+1;
		  *(p++) = i;
		  *(p++) = j+1;
		  *(p++) = k+1;
		}
	  }
  for( int i=bbxm; i<bbxp; i++ ){
	  for( int j=bbym; j<bbyp; j++ ){
		  vl[i][j] += bbzm;
		  delete[] vl[i][j];
	  }
	  vl[i] += bbym;
	  delete[] vl[i];
  }
  vl += bbxm;
  delete[] vl;
}

void
Glarea:: colourize( float& a ) {
	if( header.type() == IGB_RGBA ) {
	  glColor4ubv( (GLubyte *)(&a) );
	} else
	  cscale->colourize( a );
}

