/*
	flounder - copyright 2002 Edward J. Vigmond

	This file is part of flounder.     
	
	flounder is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.     
	
	flounder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.     
	You should have received a copy of the GNU General Public License
    along with flounder; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* isovis.h */

/* This program implements the marching cubes surface tiler described by
 * Lorensen & Cline in the Siggraph 87 Conference Proceedings.
 *
 * Original program Written by Mike Krogh, NCSA, Feb.  2, 1990
 *
 * Portions developed at the National Center for Supercomputing Applications at
 * the University of Illinois at Urbana-Champaign.
 *
 * modified by Brian Tierney, LBL  12/90
 */

/*   This program is copyright (C) 1990, Regents  of  the
University  of  California.   Anyone may reproduce this software,
in whole or in part, provided that:
(1)  Any copy  or  redistribution  must  show  the
     Regents  of  the  University of California, through its
     Lawrence Berkeley Laboratory, as the source,  and  must
     include this notice;
(2)  Any use of this software must reference this  distribu-
     tion,  state that the software copyright is held by the
     Regents of the University of California, and  that  the
     software is used by their permission.

     It is acknowledged that the U.S. Government has  rights
to this software under  Contract DE-AC03-765F00098 between the U.S.
Department of Energy and the University of California.

     This software is provided as a professional  academic  contribu-
tion  for  joint exchange.  Thus it is experimental, is pro-
vided ``as is'', with no warranties of any kind  whatsoever,
no  support,  promise  of updates, or printed documentation.
Bug reports or fixes may be sent to the author, who may or may
not act on them as he desires.
*/

/*   Author:  Brian L. Tierney
 *            Lawrence Berkeley Laboratory
 *            Imaging Technologies Group
 *            email: bltierney@lbl.gov
*/


#include <stdio.h>
#include <sys/types.h>
#ifdef CRAY
#include <sys/times.h>
#else
#include <sys/time.h>
#include <sys/timeb.h>
#endif

/* #define DEBUG */
/* #define TESTING */

// #define SPEED                   /* show execution speed */

#ifdef CRAY
#define MEM_SHORTAGE
#define CHUNK_SIZE 7		/* NOTE: minimum chunk size = 3  */
#else
#define CHUNK_SIZE 80		/* NOTE: minimum chunk size = 3  */
#endif

 /*
  * CHUNK_SIZE is the number of slices to process before writing out the
  * vertices that have been accumulated so far. If you have lots of memory,
  * might as well set this to a large number like 100
  */

#define LARGE_INT 100000000
#define MAX_DIM 512		/* maximum dimension of input data set */


#define Fread(a,b,c,d) fread((char *)(a), b, (int)(c), d)
#define Calloc(x,y) (y *)calloc((unsigned)(x), sizeof(y))
#define Fwrite(a,b,c,d) fwrite((char *)(a), b, (int)(c), d)
#define Memset(a,b,c) memset((char *)(a), b, (int)(c))

typedef struct {		/* used to detect duplicate vertices */
    int       edge_index;
    int       num_edges;
}         CUBE_TRIANGLES;

typedef struct {
    int       nverts;		/* possible values: 0-7 */
    int       verts[8];
    int       nedges;		/* possible values: 0-12 */
    int       edges[12];
    int       npolys;		/* possible values: 0-10 */
    int       polys[30];	/* 10 polys * 3 vertices = 30 */
}         CELL_ENTRY;

 /*
  * type of char is all that is necessary, but int should be much faster on
  * the Cray
  */
#ifdef CRAY
typedef int Grid_type;		/* element type for segmentation grid array */
#else
typedef char Grid_type;
#endif

typedef struct s_item {		/* stack used to simulate recursion for
				 * segmentation */
    short     i, j, k;
}         STACK_ITEM;

namespace Isosurf {

#ifdef MAIN
/* set command line arguement defaults */
int       DO_NORMALS = 0;	/* compute vertex normals flag */
int       VERBOSE = 0;		/* print status messages */
int       DUP_CHECK = 0;	/* check for duplicate vertices */
int       OUTPUT_TYPE = 0;	/* type of output file: XGL, SCRY, or ASCII */
int       SEG_METHOD = 0;	/* data segmentation method: threshold (0) or
				 * flood fill (1) */
int       DO_BRIDGES = 1;	/* in flood_fill segmentation: check strength
				 * of connecting bridges before continuing
				 * the fill */
int       SKIP_HOLES = 1;	/* this flag enables code to look for holes
				 * while thresholding, and then skip those
				 * holes while tessellating. */
int       SMALLER_BOX = 0;	/* flag indicating that the entire data set
				 * is not being examined */
int       SX = 0, SY = 0, SZ = 0;	/* starting and ending locations to
					 * examine from the original data
					 * set. */
int       EX = 9999, EY = 9999, EZ = 9999;
float     DIFF = 0.0;		/* combine vertices if difference between
				 * them is less than this value (must be used
				 * in conjunction with DUP_CHECK */

float  ***data = NULL;		/* hold input data */

int             num_vertices = 0, num_conn = 0;
float 		   	*xvertices = NULL;
float    		*yvertices = NULL;
float  		 	*zvertices = NULL;
float   		*xnormals = NULL;
float    		*ynormals = NULL;
float   		*znormals = NULL;
int     		*connectivity = NULL;

#else
extern int       DO_NORMALS ;
extern int       VERBOSE ;
extern int       DUP_CHECK ;
extern int       OUTPUT_TYPE ;
extern int       SEG_METHOD ;
extern int       DO_BRIDGES ;
extern int       SKIP_HOLES ;
extern int       SMALLER_BOX ;
extern int       SX ;
extern int       EX ;
extern float     DIFF ;

extern float  ***data;		/* hold input data */

extern int           num_vertices, num_conn;
extern float 	   	*xvertices;
extern float   		*yvertices;
extern float  	 	*zvertices;
extern float   		*xnormals;
extern float  		*ynormals;
extern float   		*znormals;
extern int     		*connectivity;
#endif

// allocate a 3D matrix to store values
void alloc_3d_data_array(int nx, int ny, int nz, float *dp=NULL );
// free the above allocated matrix
void free_3d_data_array(int keep_data );
// output the polygon info
void dump_chunk_polys( FILE *, int more, int output_type );
// main routine to find isosurface
int  make_isosurface( int xd, int yd, int zd, float thresh );
}
